/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

#define PRIVATE static

typedef enum
{
  E_FALSE = 0,
  E_TRUE = 1
} TE_BOOL;

typedef struct _TS_MAP
{
  char *pc_mapname;
  char *pc_description;
  struct _TS_MAP *pts_next;
} TS_MAP;


PRIVATE TE_BOOL r_parse_line( char *pc_in,
                              /*@out@*/ char **ppc_out_1,
                              /*@out@*/ char **ppc_out_2,
                              /*@out@*/ char **ppc_out_comment );

PRIVATE void r_get_input( char *pc_maplist,
                          /*@out@*/ TS_MAP **ppts_map_list,
                          /*@out@*/ int *pi_map_count );

PRIVATE void r_create_output( TS_MAP *pts_map_list,
                              int i_map_count,
                              char *pc_maplist,
                              char *pc_cfg );


PRIVATE TE_BOOL r_parse_line( char *pc_in,
                              char **ppc_out_1,
                              char **ppc_out_2,
                              char **ppc_out_comment )
{
  TE_BOOL b_result;

  b_result = E_TRUE;
  *ppc_out_1 = NULL;
  *ppc_out_2 = NULL;
  *ppc_out_comment = NULL;

  while (
          ( *pc_in == ' ' ) ||
          ( *pc_in == '\t' )
        )
  {
    pc_in++;
  }

  if ( *pc_in == '\0' )
  {
    // Empty line
    return E_TRUE;
  }

  if ( strncmp( pc_in, "//", 2 ) == 0 )
  {
    // Only comment
    *ppc_out_comment = pc_in;
    return E_TRUE;
  }

  // Get first string
  if ( *pc_in == '"' )
  {
    pc_in++;
    *ppc_out_1 = pc_in;

    while (
            ( *pc_in != '\0' ) &&
            ( *pc_in != '"' )
          )
    {
      pc_in++;
    }
    if ( *pc_in == '"' )
    {
      *pc_in = '\0';
      pc_in++;
    }
    else
    {
      // No closing quote
      return E_FALSE;
    }
  }
  else if ( *pc_in != '\0' )
  {
    *ppc_out_1 = pc_in;

    while (
            ( *pc_in != '\0' ) &&
            ( *pc_in != ' ' ) &&
            ( *pc_in != '\t' )
          )
    {
      pc_in++;
    }
    if (
         ( *pc_in == ' ' ) ||
         ( *pc_in == '\t' )
       )
    {
      *pc_in = '\0';
      pc_in++;
    }
    else
    {
      // We need 2 strings
      return E_FALSE;
    }
  }

  // Get second string
  while (
          ( *pc_in == ' ' ) ||
          ( *pc_in == '\t' )
        )
  {
    pc_in++;
  }

  if ( strncmp( pc_in, "//", 2 ) == 0 )
  {
    // Comment, but expected second string
    *ppc_out_comment = pc_in;
    return E_FALSE;
  }

  if ( *pc_in == '"' )
  {
    pc_in++;
    *ppc_out_2 = pc_in;

    while (
            ( *pc_in != '\0' ) &&
            ( *pc_in != '"' )
          )
    {
      pc_in++;
    }
    if ( *pc_in == '"' )
    {
      *pc_in = '\0';
      pc_in++;
    }
    else
    {
      // No closing quote
      return E_FALSE;
    }
  }
  else if ( *pc_in != '\0' )
  {
    *ppc_out_2 = pc_in;

    while (
            ( *pc_in != '\0' ) &&
            ( *pc_in != ' ' ) &&
            ( *pc_in != '\t' )
          )
    {
      pc_in++;
    }
    if (
         ( *pc_in == ' ' ) ||
         ( *pc_in == '\t' )
       )
    {
      *pc_in = '\0';
      pc_in++;
    }
    else
    {
      // Is ok here...
    }
  }

  while (
          ( *pc_in == ' ' ) ||
          ( *pc_in == '\t' )
        )
  {
    pc_in++;
  }

  if ( *pc_in != '\0' )
  {
    if ( strncmp( pc_in, "//", 2 ) == 0 )
    {
      *ppc_out_comment = pc_in;
    }
    else
    {
      // Unknown
      return E_FALSE;
    }
  }

  return E_TRUE;
}

#define D_BUFFER_SIZE 1024

// Read input file
PRIVATE void r_get_input( char *pc_maplist,
                          TS_MAP **ppts_map_list,
                          int *pi_map_count )
{
  FILE *po_file;
  TS_MAP *pts_ptr;
  char ac_line[D_BUFFER_SIZE];
  char *pc_result;
  TE_BOOL b_result;
  char *pc_mapname;
  char *pc_description;
  char *pc_comment;
  char *pc_ptr;
  size_t o_len;

  *ppts_map_list = NULL;
  *pi_map_count = 0;

  // Open map input file
  po_file = fopen( pc_maplist, "rt" );
  if ( po_file == NULL )
  {
    fprintf( stderr, "File open error: %s\n", pc_maplist );
    exit( EXIT_FAILURE );
  }

  // Read first line
  pc_result = fgets( ac_line, D_BUFFER_SIZE, po_file );
  while ( pc_result != NULL )
  {
    // Process line
    o_len = strlen( ac_line );
    if (
         ( o_len > 0 ) &&
         ( ac_line[o_len - 1] == '\n' )
       )
    {
      ac_line[o_len - 1] = '\0';
    }

    b_result = r_parse_line( ac_line,
                             &pc_mapname,
                             &pc_description,
                             &pc_comment );

    if ( b_result == E_FALSE )
    {
      (void)fprintf( stderr, "Invalid input file\n" );
      exit( EXIT_FAILURE );
    }

    if (
         ( pc_mapname != NULL ) &&
         ( pc_description != NULL )
       )
    {
      pts_ptr = (TS_MAP *) malloc( sizeof(TS_MAP) );
      if ( pts_ptr == NULL )
      {
        (void)fprintf( stderr, "Memory allocation error\n" );
        exit( EXIT_FAILURE );
      }

      // Store mapname (replace '\' by '/' and convert to lowercase)
      o_len = strlen( pc_mapname );
      pts_ptr->pc_mapname = (char *) malloc( ( o_len + 1 ) * sizeof(char) );
      if ( pts_ptr->pc_mapname == NULL )
      {
        (void)fprintf( stderr, "Memory allocation error\n" );
        exit( EXIT_FAILURE );
      }
      pc_ptr = pts_ptr->pc_mapname;
      while ( *pc_mapname != '\0' )
      {
        if ( *pc_mapname == '\\' )
        {
          *pc_ptr = '/';
        }
        else
        {
          *pc_ptr = tolower( *pc_mapname );
        }

        pc_ptr++;
        pc_mapname++;
      }

      // Store description
      o_len = strlen( pc_description );
      pts_ptr->pc_description = (char *) malloc( ( o_len + 1 ) * sizeof(char) );
      if ( pts_ptr->pc_description == NULL )
      {
        (void)fprintf( stderr, "Memory allocation error\n" );
        exit( EXIT_FAILURE );
      }
      (void)memcpy( pts_ptr->pc_description, pc_description, o_len + 1 );

      // Add to list
      pts_ptr->pts_next = *ppts_map_list;
      *ppts_map_list = pts_ptr;
      (*pi_map_count)++;
    }

    // Read next line
    pc_result = fgets( ac_line, D_BUFFER_SIZE, po_file );
  }

  return;
}

// Create output files
PRIVATE void r_create_output( TS_MAP *pts_map_list,
                              int i_map_count,
                              char *pc_maplist,
                              char *pc_cfg )
{
  FILE *po_file;
  TS_MAP **ppts_map_arr;
  TS_MAP *pts_ptr;
  int i_rand;
  char *pc_first_map;
  int i_count;

  // Fill array
  ppts_map_arr = (TS_MAP **) malloc( sizeof(TS_MAP *) * i_map_count );
  if ( ppts_map_arr == NULL )
  {
    (void)fprintf( stderr, "Memory allocation error\n" );
    exit( EXIT_FAILURE );
  }
  pts_ptr = pts_map_list;
  for ( i_count = 0; i_count < i_map_count; i_count++ )
  {
    ppts_map_arr[i_count] = pts_ptr;
    pts_ptr = pts_ptr->pts_next;
  }

  // Create map output file
  po_file = fopen( pc_maplist, "wt" );
  if ( po_file == NULL )
  {
    (void)fprintf( stderr, "File create error: %s\n", pc_maplist );
    exit( EXIT_FAILURE );
  }

  // Initialize random number generator
  srand( (unsigned int)time( NULL ) );
  (void)rand();

  pc_first_map = NULL;
  i_count = i_map_count;
  while ( i_count >= 1 )
  {
    // Get random position (chances aren't equal, but it's good enough)
    i_rand = rand() % i_count;

    // Write map to file
    (void)fprintf( po_file, "%s\t\"%s\"\n", ppts_map_arr[i_rand]->pc_mapname, ppts_map_arr[i_rand]->pc_description );

    // Remember first map
    if ( pc_first_map == NULL )
    {
      pc_first_map = ppts_map_arr[i_rand]->pc_mapname;
    }

    // Remove used map from array
    i_count--;
    if ( i_rand < i_count )
    {
      (void)memmove( &ppts_map_arr[i_rand], &ppts_map_arr[i_rand + 1], ( i_count - i_rand ) * sizeof(TS_MAP *) );
    }

  }

  // Close map output file
  (void)fclose( po_file );  


  // Write to cfg if requested
  if ( pc_cfg != NULL )
  {
    // Create cfg output file
    po_file = fopen( pc_cfg, "wt" );
    if ( po_file == NULL )
    {
      (void)fprintf( stderr, "File create error: %s\n", pc_cfg );
      exit( EXIT_FAILURE );
    }

    // Write first map name to file
    fprintf( po_file, "set sv_first_map \"%s\"\n", pc_first_map );

    // Close cfg output file
    (void)fclose( po_file );
  }


  free( ppts_map_arr );

  return;
}

int main( int argc, char *argv[] )
{
  TS_MAP *pts_map_list;
  int i_map_count;
  TS_MAP *pts_ptr;

  if (
       ( argc < 3 ) ||
       ( argc > 4 )
     )
  {
    (void)fprintf( stderr, "Invalid number of arguments\n"
                           "\n"
                           "Usage:\n"
                           "  %s <maplist in> <maplist out> [cfg out]\n",
                           argv[0] );
     return EXIT_FAILURE;
  }

  r_get_input( argv[1],
               &pts_map_list,
               &i_map_count );

  if ( argc > 3 )
  {
    r_create_output( pts_map_list,
                     i_map_count,
                     argv[2],
                     argv[3] );
  }
  else
  {
    r_create_output( pts_map_list,
                     i_map_count,
                     argv[2],
                     NULL );
  }


  while ( pts_map_list != NULL )
  {
    free( pts_map_list->pc_mapname );
    free( pts_map_list->pc_description );
    pts_ptr = pts_map_list->pts_next;
    free( pts_map_list );
    pts_map_list = pts_ptr;
  }

  return EXIT_SUCCESS;
}
